#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#include <dirent.h>
#include <stdint.h>
#include <sys/wait.h>
#include "fsig.h"


#define HDR_EXPLAIN \
"This is the INDEX file for an image TAR. The structure of this file is \n" \
"as follows:\n" \
" - This text\n" \
" - Some optional space\n" \
" - Magic bytes\n" \
" - Number of offsets (8 bytes, little endian)\n" \
" - Offset table (8 bytes each, little endian)\n" \
"\n" \
"Each entry in the offset table corresponds to an absolute offset from the\n" \
"start of the TAR file. The offset points to the actual contents of the \n" \
"enclosed file, not the TAR header or anything like that.\n\n" \
"This file is supposed to go at the beginning of the TAR file so that the\n" \
"dumb readers can quickly search for the magic bytes.\n\n"


/* The program puts the given files into the given TAR using the tar 
   command. After that, it searches for each file in the resulting TAR
   and records the offsets. These are then written to the INDEX file
   inside the TAR. 

   The first command line argument can be 'c' or 'u'. c stands for create
   and 'u' stands for update. 

   The second should be the name of the TAR file.

   The rest are the files to be inserted into the TAR file. When creating,
   this list is prefixed with the file INDEX, which is created in the current
   directory. */

FILE *open_force(char *name, char *mode)
{
  FILE *f;
  f= fopen(name,mode);
  if (!f) exit(fprintf(stderr,"can't open file %s in mode %s: %s\n", 
                               name, mode, strerror(errno)));
  return f;
}

static void putu64(char *out, uint64_t V)
{
  int i;
  for(i=0;i<8;i++) 
  {
    out[i]= V;
    V>>= 8;
  }
}

static void write_bytes(FILE *f, char *data, int len)
{
  while(len)
  {
    int R;
    R= fwrite(data, 1, len, f);
    if (R<=0) exit(fprintf(stderr,"can't write %d bytes to file: %s\n", len,
                 strerror(errno)));
    len-= R;
    data+= R;
  }
}

static void read_sig(char *fn, char *sig, int siz, int *Rlen)
{
  FILE *f;
  f= open_force(fn, "rb");
  *Rlen= fread(sig, 1,siz, f);
  fclose(f);
}

static void usage(char **argv)
{
  exit(fprintf(stderr,"usage:\n"
         "  %s file.tar input1.png input2.png ...\n"
         "      creates the given TAR file with index.\n", argv[0]));
}


int main(int argc,char **argv)
{
  char *xpla, *head;
  int nfiles;
  int i;
  FILE *f;
  char *tarname;
  char *table;
  int indexsz;
  uint64_t hpos;
  char *index;

    char **targs;

  pid_t pid;
  int chstatus;

  if (argc<4) usage(argv);


  tarname= argv[1];

  nfiles= argc - 2;
  xpla= HDR_EXPLAIN;
  head= HDR_MAGIC;
  indexsz= strlen(xpla) + strlen(head) + 8 + nfiles*8;
  index= calloc(1,indexsz);


    targs= malloc(sizeof(*targs)*(nfiles+5));
    i= 0;
    targs[0]= "tar";
    targs[1]= "cf";
    targs[2]= tarname;
    targs[3]= "INDEX";
    for(i=2;i<argc;i++) targs[i-2+4]= argv[i];
    targs[i-2+4]= NULL;

    f= open_force("INDEX","wb");
    int offset,L;
    offset= 0;
    L= strlen(xpla); memcpy(index+offset, xpla, L); offset+= L;
    L= strlen(head); memcpy(index+offset, head, L); offset+= L;

    putu64(index+offset, nfiles);

    write_bytes(f, index, indexsz);
    fclose(f);
    printf("INDEX file created, running tar.\n");

    pid= fork();
    if (pid<0) exit(fprintf(stderr,"can not fork: %s\n",strerror(errno)));
    if (pid)
    {
      pid= waitpid(pid, &chstatus, 0);
      if (pid<0) exit(fprintf(stderr,"waitpid failed: %s\n",strerror(errno)));
      if (!WIFEXITED(chstatus) || WEXITSTATUS(chstatus)!=0)
        exit(fprintf(stderr,"child exited abnormally, exiting.\n"));
    }
    else
      execvp("tar", targs);
 

  f= open_force(tarname,"r+b");

  uint64_t *pos;
  pos= malloc(sizeof(*pos)*(nfiles+1));

  for(i=0;i<nfiles;i++)
  {
    char sig[1024];
    int siglen;
    read_sig(argv[i+2], sig, 1024, &siglen);

    pos[i+1]= find_sig(f, sig, siglen);
    if (!pos[i+1]) 
       exit(fprintf(stderr,"can't find file %s in TAR.\n", argv[i+2]));
  }
  pos[0]= nfiles;

  fseek(f, 0, SEEK_SET);
  hpos= find_sig(f, head, strlen(head));
  if (!hpos) 
       exit(fprintf(stderr,"can't find IMGTAR magic in the TAR file.\n"));

  fseek(f, hpos+strlen(head), SEEK_SET);
  table= malloc(8*(nfiles+1));
  for(i=0;i<=nfiles;i++)
   putu64(table+i*8, pos[i]);

  write_bytes(f, table, (nfiles+1)*8);
  printf("index written, all complete.\n");
  fclose(f);
  return 0;
}

