/***********************************************************************
 *                                                                     *
 *                               Decoder                               *
 *                                                                     *
 ***********************************************************************
 ***********************************************************************/
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <zlib.h>
#include <stdarg.h>
#include "pngio.h"
#include "firead.h"
#include "../encode/fsig.h"


struct ipak
{
  FILE *f;
  char *error;
  int npages;
  uint64_t *index;
};

static int im_error(ipak_t *IM, const char *fmt, ...)
{
  char buf[2000];
  va_list ap;
  va_start(ap,fmt);
  vsnprintf(buf, sizeof(buf)-1, fmt, ap);
  va_end(ap);
  buf[sizeof(buf)-1]= 0;
  IM->error= strdup(buf);
  return 1;
}

static int set_fpos(ipak_t *IM, unsigned int pos)
{
  if (fseek(IM->f, pos, SEEK_SET))
    return im_error(IM,"can't seek in output file: %s\n",strerror(errno));
  return 0;
}

static int read_bytes(ipak_t *IM, void *dst, int len)
{
  int R;
  R= fread(dst,1,len,IM->f);
  if (R!=len)
    return im_error(IM,"error during read");
  return 0;
}



int ipak_render_page(ipak_t *IM,int pg, unsigned char **Rimg,
                     unsigned int *Rw,unsigned int *Rh)
{
  unsigned int W,H;
  unsigned char *img;
  char *pe;
  int eof;

  if (IM->error) return 1;
  if (pg<0 || pg>=IM->npages) return 1;

  if (set_fpos(IM, IM->index[pg])) return 1;

  pe= read_png_g8_filep(IM->f,&W,&H,&img,&eof);
  if (pe)
  {
    im_error(IM,"png error: %s",pe);
    free(pe);
    return 1;
  }

  *Rimg= img; *Rw= W; *Rh= H;
  return 0;
}

static void freenotnull(void *a) { if (a) free(a); }

void ipak_close(ipak_t *IM)
{
  if (IM->f) fclose(IM->f);

  freenotnull(IM->index);
  freenotnull(IM->error);
  memset(IM, 0, sizeof(*IM));
  free(IM);
}

char *ipak_error(ipak_t *IM)
{
  return IM->error;
}

static uint64_t dint8(uint8_t *B)
{
  uint64_t V= 0;
  uint64_t P;
  int i;
  for(i=0;i<8;i++)
  {
    P= B[i];
    P<<= i*8;
    V|= P;
  }
  return V;
}


int ipak_init(ipak_t **RIM,char *fn, int *Rnpages)
{
  ipak_t *IM;
  int mL;
  uint8_t Bnpa[8];
  uint8_t *Btab;
  uint64_t hdrpos;
  int i;

  *RIM= IM= calloc(1,sizeof(*IM));
  IM->f= fopen(fn, "rb");
  if (!IM->f) return im_error(IM,"can't open input %s: %s",fn,strerror(errno));

  mL= strlen(HDR_MAGIC);

  hdrpos= find_sig(IM->f, HDR_MAGIC, mL);

  if (!hdrpos) return im_error(IM,"can't find signature in file %s",fn);

  if (set_fpos(IM, hdrpos+mL)) return 1;
  if (read_bytes(IM, Bnpa, 8)) return 1;
  IM->npages= dint8(Bnpa);
  if (IM->npages>10000 || IM->npages<=0)
     return im_error(IM,"too big or corrupt file %s (npages=%d)",fn,IM->npages);
  *Rnpages= IM->npages;
  IM->index= malloc(sizeof(uint64_t)*IM->npages);
  Btab= malloc(IM->npages*8);
  if (read_bytes(IM, Btab, IM->npages*8))
  {
     free(Btab);
     return 1;
  }
  for(i=0;i<IM->npages;i++)
    IM->index[i]= dint8(Btab+i*8);
  free(Btab);
  return 0;
}
